from yt.units.yt_array import YTArray
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py


def save_as_dataset(ds, filename, data, field_types=None, extra_attrs=None):
    r"""Export a set of field arrays to a reloadable yt dataset.

    This function can be used to create a yt loadable dataset from a
    set of arrays.  The field arrays can either be associated with a
    loaded dataset or, if not, a dictionary of dataset attributes can
    be provided that will be used as metadata for the new dataset.  The
    resulting dataset can be reloaded as a yt dataset.

    Parameters
    ----------
    ds : dataset or dict
        The dataset associated with the fields or a dictionary of
        parameters.
    filename : str
        The name of the file to be written.
    data : dict
        A dictionary of field arrays to be saved.
    field_types: dict, optional
        A dictionary denoting the group name to which each field is to
        be saved. When the resulting dataset is reloaded, this will be
        the field type for this field. If not given, "data" will be
        used.
    extra_attrs: dict, optional
        A dictionary of additional attributes to be saved.

    Returns
    -------
    filename : str
        The name of the file that has been created.

    Examples
    --------

    >>> import numpy as np
    >>> import yt
    >>> ds = yt.load("enzo_tiny_cosmology/DD0046/DD0046")
    >>> sphere = ds.sphere([0.5] * 3, (10, "Mpc"))
    >>> sphere_density = sphere["gas", "density"]
    >>> region = ds.box([0.0] * 3, [0.25] * 3)
    >>> region_density = region["gas", "density"]
    >>> data = {}
    >>> data["sphere_density"] = sphere_density
    >>> data["region_density"] = region_density
    >>> yt.save_as_dataset(ds, "density_data.h5", data)
    >>> new_ds = yt.load("density_data.h5")
    >>> print(new_ds.data["region_density"])
    [  7.47650434e-32   7.70370740e-32   9.74692941e-32 ...,   1.22384547e-27
       5.13889063e-28   2.91811974e-28] g/cm**3
    >>> print(new_ds.data["sphere_density"])
    [  4.46237613e-32   4.86830178e-32   4.46335118e-32 ...,   6.43956165e-30
       3.57339907e-30   2.83150720e-30] g/cm**3
    >>> data = {
    ...     "density": yt.YTArray(1e-24 * np.ones(10), "g/cm**3"),
    ...     "temperature": yt.YTArray(1000.0 * np.ones(10), "K"),
    ... }
    >>> ds_data = {"current_time": yt.YTQuantity(10, "Myr")}
    >>> yt.save_as_dataset(ds_data, "random_data.h5", data)
    >>> new_ds = yt.load("random_data.h5")
    >>> print(new_ds.data["gas", "temperature"])
    [ 1000.  1000.  1000.  1000.  1000.  1000.  1000.  1000.  1000.  1000.] K

    """

    mylog.info("Saving field data to yt dataset: %s.", filename)

    if extra_attrs is None:
        extra_attrs = {}
    base_attrs = [
        "dimensionality",
        "domain_left_edge",
        "domain_right_edge",
        "current_redshift",
        "current_time",
        "domain_dimensions",
        "geometry",
        "periodicity",
        "cosmological_simulation",
        "omega_lambda",
        "omega_matter",
        "hubble_constant",
        "length_unit",
        "mass_unit",
        "time_unit",
        "velocity_unit",
        "magnetic_unit",
    ]

    fh = h5py.File(filename, mode="w")
    if ds is None:
        ds = {}

    if hasattr(ds, "parameters") and isinstance(ds.parameters, dict):
        for attr, val in ds.parameters.items():
            _yt_array_hdf5_attr(fh, attr, val)

    if hasattr(ds, "unit_registry"):
        _yt_array_hdf5_attr(fh, "unit_registry_json", ds.unit_registry.to_json())

    if hasattr(ds, "unit_system"):
        # Note: ds._unit_system_name is written here rather than
        # ds.unit_system.name because for a 'code' unit system, ds.unit_system.name
        # is a hash, not a unit system. And on re-load, we want to designate
        # a unit system not a hash value.
        # See https://github.com/yt-project/yt/issues/4315 for more background.
        _yt_array_hdf5_attr(fh, "unit_system_name", ds._unit_system_name)

    for attr in base_attrs:
        if isinstance(ds, dict):
            my_val = ds.get(attr, None)
        else:
            my_val = getattr(ds, attr, None)
        if my_val is None:
            continue
        _yt_array_hdf5_attr(fh, attr, my_val)

    for attr in extra_attrs:
        my_val = extra_attrs[attr]
        _yt_array_hdf5_attr(fh, attr, my_val)
    if "data_type" not in extra_attrs:
        fh.attrs["data_type"] = "yt_array_data"

    for field in data:
        if field_types is None:
            field_type = "data"
        else:
            field_type = field_types[field]
        if field_type not in fh:
            fh.create_group(field_type)

        if isinstance(field, tuple):
            field_name = field[1]
        else:
            field_name = field

        # for python3
        if data[field].dtype.kind == "U":
            data[field] = data[field].astype("|S")

        _yt_array_hdf5(fh[field_type], field_name, data[field])
        if "num_elements" not in fh[field_type].attrs:
            fh[field_type].attrs["num_elements"] = data[field].size
    fh.close()
    return filename


def _hdf5_yt_array(fh, field, ds=None):
    r"""Load an hdf5 dataset as a YTArray.

    Reads in a dataset from an open hdf5 file or group and uses the
    "units" attribute, if it exists, to apply units.

    Parameters
    ----------
    fh : an open hdf5 file or hdf5 group
        The hdf5 file or group in which the dataset exists.
    field : str
        The name of the field to be loaded.
    ds : yt Dataset
        If not None, the unit_registry of the dataset
        is used to apply units.

    Returns
    -------
    A YTArray of the requested field.

    """

    if ds is None:
        new_arr = YTArray
    else:
        new_arr = ds.arr
    units = ""
    if "units" in fh[field].attrs:
        units = fh[field].attrs["units"]
    if units == "dimensionless":
        units = ""
    return new_arr(fh[field][()], units)


def _yt_array_hdf5(fh, field, data):
    r"""Save a YTArray to an open hdf5 file or group.

    Save a YTArray to an open hdf5 file or group, and save the
    units to a "units" attribute.

    Parameters
    ----------
    fh : an open hdf5 file or hdf5 group
        The hdf5 file or group to which the data will be written.
    field : str
        The name of the field to be saved.
    data : YTArray
        The data array to be saved.

    Returns
    -------
    dataset : hdf5 dataset
        The created hdf5 dataset.

    """

    dataset = fh.create_dataset(str(field), data=data)
    units = ""
    if isinstance(data, YTArray):
        units = str(data.units)
    dataset.attrs["units"] = units
    return dataset


def _yt_array_hdf5_attr(fh, attr, val):
    r"""Save a YTArray or YTQuantity as an hdf5 attribute.

    Save an hdf5 attribute.  If it has units, save an
    additional attribute with the units.

    Parameters
    ----------
    fh : an open hdf5 file, group, or dataset
        The hdf5 file, group, or dataset to which the
        attribute will be written.
    attr : str
        The name of the attribute to be saved.
    val : anything
        The value to be saved.

    """

    if val is None:
        val = "None"
    if hasattr(val, "units"):
        fh.attrs[f"{attr}_units"] = str(val.units)
    try:
        fh.attrs[str(attr)] = val
    # This is raised if no HDF5 equivalent exists.
    # In that case, save its string representation.
    except TypeError:
        fh.attrs[str(attr)] = str(val)



import numpy as np
import pytest
import unyt

import yt
from yt.data_objects.selection_objects.region import YTRegion
from yt.testing import (
    assert_rel_equal,
    cubicspline_python,
    distancematrix,
    fake_random_sph_ds,
    fake_sph_flexible_grid_ds,
    integrate_kernel,
)


@pytest.mark.parametrize("weighted", [True, False])
@pytest.mark.parametrize("periodic", [True, False])
@pytest.mark.parametrize("depth", [None, (1.0, "cm")])
@pytest.mark.parametrize("shiftcenter", [False, True])
@pytest.mark.parametrize("axis", [0, 1, 2])
def test_sph_proj_general_alongaxes(
    axis: int,
    shiftcenter: bool,
    depth: float | None,
    periodic: bool,
    weighted: bool,
) -> None:
    """
    The previous projection tests were for a specific issue.
    Here, we test more functionality of the projections.
    We just send lines of sight through pixel centers for convenience.
    Particles at [0.5, 1.5, 2.5] (in each coordinate)
    smoothing lengths 0.25
    all particles have mass 1., density 1.5,
    except the single center particle, with mass 2., density 3.

    Parameters:
    -----------
    axis: {0, 1, 2}
        projection axis (aligned with sim. axis)
    shiftcenter: bool
        shift the coordinates to center the projection on.
        (The grid is offset to this same center)
    depth: float or None
        depth of the projection slice
    periodic: bool
        assume periodic boundary conditions, or not
    weighted: bool
        make a weighted projection (density-weighted density), or not

    Returns:
    --------
    None
    """
    if shiftcenter:
        center = unyt.unyt_array(np.array((0.625, 0.625, 0.625)), "cm")
    else:
        center = unyt.unyt_array(np.array((1.5, 1.5, 1.5)), "cm")
    bbox = unyt.unyt_array(np.array([[0.0, 3.0], [0.0, 3.0], [0.0, 3.0]]), "cm")
    hsml_factor = 0.5
    unitrho = 1.5

    # test correct centering, particle selection
    def makemasses(i, j, k):
        if i == j == k == 1:
            return 2.0
        else:
            return 1.0

    # m / rho, factor 1. / hsml**2 is included in the kernel integral
    # (density is adjusted, so same for center particle)
    prefactor = 1.0 / unitrho  # / (0.5 * 0.5)**2
    dl_cen = integrate_kernel(cubicspline_python, 0.0, 0.25)

    # result shouldn't depend explicitly on the center if we re-center
    # the data, unless we get cut-offs in the non-periodic case
    ds = fake_sph_flexible_grid_ds(
        hsml_factor=hsml_factor,
        nperside=3,
        periodic=periodic,
        offsets=np.full(3, 0.5),
        massgenerator=makemasses,
        unitrho=unitrho,
        bbox=bbox.v,
        recenter=center.v,
    )
    if depth is None:
        source = ds.all_data()
    else:
        depth = unyt.unyt_quantity(*depth)
        le = np.array(ds.domain_left_edge)
        re = np.array(ds.domain_right_edge)
        le[axis] = center[axis] - 0.5 * depth
        re[axis] = center[axis] + 0.5 * depth
        cen = 0.5 * (le + re)
        reg = YTRegion(center=cen, left_edge=le, right_edge=re, ds=ds)
        source = reg

    # we don't actually want a plot, it's just a straightforward,
    # common way to get an frb / image array
    if weighted:
        toweight_field = ("gas", "density")
    else:
        toweight_field = None
    prj = yt.ProjectionPlot(
        ds,
        axis,
        ("gas", "density"),
        width=(2.5, "cm"),
        weight_field=toweight_field,
        buff_size=(5, 5),
        center=center,
        data_source=source,
    )
    img = prj.frb.data[("gas", "density")]
    if weighted:
        expected_out = np.zeros(
            (
                5,
                5,
            ),
            dtype=img.v.dtype,
        )
        expected_out[::2, ::2] = unitrho
        if depth is None:
            ## during shift, particle coords do wrap around edges
            # if (not periodic) and shiftcenter:
            #    # weight 1. for unitrho, 2. for 2. * untrho
            #    expected_out[2, 2] *= 5. / 3.
            # else:
            # weight (2 * 1.) for unitrho, (1 * 2.) for 2. * unitrho
            expected_out[2, 2] *= 1.5
        else:
            # only 2 * unitrho element included
            expected_out[2, 2] *= 2.0
    else:
        expected_out = np.zeros(
            (
                5,
                5,
            ),
            dtype=img.v.dtype,
        )
        expected_out[::2, ::2] = dl_cen * prefactor * unitrho
        if depth is None:
            # 3 particles per l.o.s., including the denser one
            expected_out *= 3.0
            expected_out[2, 2] *= 4.0 / 3.0
        else:
            # 1 particle per l.o.s., including the denser one
            expected_out[2, 2] *= 2.0
    # grid is shifted to the left -> 'missing' stuff at the left
    if (not periodic) and shiftcenter:
        expected_out[:1, :] = 0.0
        expected_out[:, :1] = 0.0
    # print(axis, shiftcenter, depth, periodic, weighted)
    # print(expected_out)
    # print(img.v)
    assert_rel_equal(expected_out, img.v, 5)


@pytest.mark.parametrize("periodic", [True, False])
@pytest.mark.parametrize("shiftcenter", [False, True])
@pytest.mark.parametrize("zoff", [0.0, 0.1, 0.5, 1.0])
@pytest.mark.parametrize("axis", [0, 1, 2])
def test_sph_slice_general_alongaxes(
    axis: int,
    shiftcenter: bool,
    periodic: bool,
    zoff: float,
) -> None:
    """
    Particles at [0.5, 1.5, 2.5] (in each coordinate)
    smoothing lengths 0.25
    all particles have mass 1., density 1.5,
    except the single center particle, with mass 2., density 3.

    Parameters:
    -----------
    axis: {0, 1, 2}
        projection axis (aligned with sim. axis)
    northvector: tuple
        y-axis direction in the final plot (direction vector)
    shiftcenter: bool
        shift the coordinates to center the projection on.
        (The grid is offset to this same center)
    zoff: float
        offset of the slice plane from the SPH particle center plane
    periodic: bool
        assume periodic boundary conditions, or not

    Returns:
    --------
    None
    """
    if shiftcenter:
        center = unyt.unyt_array(np.array((0.625, 0.625, 0.625)), "cm")
    else:
        center = unyt.unyt_array(np.array((1.5, 1.5, 1.5)), "cm")
    bbox = unyt.unyt_array(np.array([[0.0, 3.0], [0.0, 3.0], [0.0, 3.0]]), "cm")
    hsml_factor = 0.5
    unitrho = 1.5

    # test correct centering, particle selection
    def makemasses(i, j, k):
        if i == j == k == 1:
            return 2.0
        elif i == j == k == 2:
            return 3.0
        else:
            return 1.0

    # result shouldn't depend explicitly on the center if we re-center
    # the data, unless we get cut-offs in the non-periodic case
    ds = fake_sph_flexible_grid_ds(
        hsml_factor=hsml_factor,
        nperside=3,
        periodic=periodic,
        offsets=np.full(3, 0.5),
        massgenerator=makemasses,
        unitrho=unitrho,
        bbox=bbox.v,
        recenter=center.v,
    )
    ad = ds.all_data()
    # print(ad[('gas', 'position')])
    outgridsize = 10
    width = 2.5
    _center = center.to("cm").v.copy()
    _center[axis] += zoff

    # we don't actually want a plot, it's just a straightforward,
    # common way to get an frb / image array
    slc = yt.SlicePlot(
        ds,
        axis,
        ("gas", "density"),
        width=(width, "cm"),
        buff_size=(outgridsize,) * 2,
        center=(_center, "cm"),
    )
    img = slc.frb.data[("gas", "density")]

    # center is same in non-projection coords
    if axis == 0:
        ci = 1
    else:
        ci = 0
    gridcens = (
        _center[ci]
        - 0.5 * width
        + 0.5 * width / outgridsize
        + np.arange(outgridsize) * width / outgridsize
    )
    xgrid = np.repeat(gridcens, outgridsize)
    ygrid = np.tile(gridcens, outgridsize)
    zgrid = np.full(outgridsize**2, _center[axis])
    gridcoords = np.empty((outgridsize**2, 3), dtype=xgrid.dtype)
    if axis == 2:
        gridcoords[:, 0] = xgrid
        gridcoords[:, 1] = ygrid
        gridcoords[:, 2] = zgrid
    elif axis == 0:
        gridcoords[:, 0] = zgrid
        gridcoords[:, 1] = xgrid
        gridcoords[:, 2] = ygrid
    elif axis == 1:
        gridcoords[:, 0] = ygrid
        gridcoords[:, 1] = zgrid
        gridcoords[:, 2] = xgrid
    ad = ds.all_data()
    sphcoords = np.array(
        [
            (ad[("gas", "x")]).to("cm"),
            (ad[("gas", "y")]).to("cm"),
            (ad[("gas", "z")]).to("cm"),
        ]
    ).T
    # print("sphcoords:")
    # print(sphcoords)
    # print("gridcoords:")
    # print(gridcoords)
    dists = distancematrix(
        gridcoords,
        sphcoords,
        periodic=(periodic,) * 3,
        periods=np.array([3.0, 3.0, 3.0]),
    )
    # print("dists <= 1:")
    # print(dists <= 1)
    sml = (ad[("gas", "smoothing_length")]).to("cm")
    normkern = cubicspline_python(dists / sml.v[np.newaxis, :])
    sphcontr = normkern / sml[np.newaxis, :] ** 3 * ad[("gas", "mass")]
    contsum = np.sum(sphcontr, axis=1)
    sphweights = (
        normkern
        / sml[np.newaxis, :] ** 3
        * ad[("gas", "mass")]
        / ad[("gas", "density")]
    )
    weights = np.sum(sphweights, axis=1)
    nzeromask = np.logical_not(weights == 0)
    expected = np.zeros(weights.shape, weights.dtype)
    expected[nzeromask] = contsum[nzeromask] / weights[nzeromask]
    expected = expected.reshape((outgridsize, outgridsize))
    # expected[np.isnan(expected)] = 0.0  # convention in the slices

    # print("expected:\n", expected)
    # print("recovered:\n", img.v)
    assert_rel_equal(expected, img.v, 5)


@pytest.mark.parametrize("periodic", [True, False])
@pytest.mark.parametrize("shiftcenter", [False, True])
@pytest.mark.parametrize("northvector", [None, (1.0e-4, 1.0, 0.0)])
@pytest.mark.parametrize("zoff", [0.0, 0.1, 0.5, 1.0])
def test_sph_slice_general_offaxis(
    northvector: tuple[float, float, float] | None,
    shiftcenter: bool,
    zoff: float,
    periodic: bool,
) -> None:
    """
    Same as the on-axis slices, but we rotate the basis vectors
    to test whether roations are handled ok. the rotation is chosen to
    be small so that in/exclusion of particles within bboxes, etc.
    works out the same way.
    Particles at [0.5, 1.5, 2.5] (in each coordinate)
    smoothing lengths 0.25
    all particles have mass 1., density 1.5,
    except the single center particle, with mass 2., density 3.

    Parameters:
    -----------
    northvector: tuple
        y-axis direction in the final plot (direction vector)
    shiftcenter: bool
        shift the coordinates to center the projection on.
        (The grid is offset to this same center)
    zoff: float
        offset of the slice plane from the SPH particle center plane
    periodic: bool
        assume periodic boundary conditions, or not

    Returns:
    --------
    None
    """
    if shiftcenter:
        center = np.array((0.625, 0.625, 0.625))  # cm
    else:
        center = np.array((1.5, 1.5, 1.5))  # cm
    bbox = unyt.unyt_array(np.array([[0.0, 3.0], [0.0, 3.0], [0.0, 3.0]]), "cm")
    hsml_factor = 0.5
    unitrho = 1.5

    # test correct centering, particle selection
    def makemasses(i, j, k):
        if i == j == k == 1:
            return 2.0
        else:
            return 1.0

    # try to make sure dl differences from periodic wrapping are small
    epsilon = 1e-4
    projaxis = np.array([epsilon, 0.00, np.sqrt(1.0 - epsilon**2)])
    e1dir = projaxis / np.sqrt(np.sum(projaxis**2))
    if northvector is None:
        e2dir = np.array([0.0, 1.0, 0.0])
    else:
        e2dir = np.asarray(northvector)
    e2dir = e2dir - np.sum(e1dir * e2dir) * e2dir  # orthonormalize
    e2dir /= np.sqrt(np.sum(e2dir**2))
    e3dir = np.cross(e2dir, e1dir)

    outgridsize = 10
    width = 2.5
    _center = center.copy()
    _center += zoff * e1dir

    ds = fake_sph_flexible_grid_ds(
        hsml_factor=hsml_factor,
        nperside=3,
        periodic=periodic,
        offsets=np.full(3, 0.5),
        massgenerator=makemasses,
        unitrho=unitrho,
        bbox=bbox.v,
        recenter=center,
        e1hat=e1dir,
        e2hat=e2dir,
        e3hat=e3dir,
    )

    # source = ds.all_data()
    # couple to dataset -> right unit registry
    center = ds.arr(center, "cm")
    # print("position:\n", source["gas", "position"])
    slc = yt.SlicePlot(
        ds,
        e1dir,
        ("gas", "density"),
        width=(width, "cm"),
        buff_size=(outgridsize,) * 2,
        center=(_center, "cm"),
        north_vector=e2dir,
    )
    img = slc.frb.data[("gas", "density")]

    # center is same in x/y (e3dir/e2dir)
    gridcenx = (
        np.dot(_center, e3dir)
        - 0.5 * width
        + 0.5 * width / outgridsize
        + np.arange(outgridsize) * width / outgridsize
    )
    gridceny = (
        np.dot(_center, e2dir)
        - 0.5 * width
        + 0.5 * width / outgridsize
        + np.arange(outgridsize) * width / outgridsize
    )
    xgrid = np.repeat(gridcenx, outgridsize)
    ygrid = np.tile(gridceny, outgridsize)
    zgrid = np.full(outgridsize**2, np.dot(_center, e1dir))
    gridcoords = (
        xgrid[:, np.newaxis] * e3dir[np.newaxis, :]
        + ygrid[:, np.newaxis] * e2dir[np.newaxis, :]
        + zgrid[:, np.newaxis] * e1dir[np.newaxis, :]
    )
    # print("gridcoords:")
    # print(gridcoords)
    ad = ds.all_data()
    sphcoords = np.array(
        [
            (ad[("gas", "x")]).to("cm"),
            (ad[("gas", "y")]).to("cm"),
            (ad[("gas", "z")]).to("cm"),
        ]
    ).T
    dists = distancematrix(
        gridcoords,
        sphcoords,
        periodic=(periodic,) * 3,
        periods=np.array([3.0, 3.0, 3.0]),
    )
    sml = (ad[("gas", "smoothing_length")]).to("cm")
    normkern = cubicspline_python(dists / sml.v[np.newaxis, :])
    sphcontr = normkern / sml[np.newaxis, :] ** 3 * ad[("gas", "mass")]
    contsum = np.sum(sphcontr, axis=1)
    sphweights = (
        normkern
        / sml[np.newaxis, :] ** 3
        * ad[("gas", "mass")]
        / ad[("gas", "density")]
    )
    weights = np.sum(sphweights, axis=1)
    nzeromask = np.logical_not(weights == 0)
    expected = np.zeros(weights.shape, weights.dtype)
    expected[nzeromask] = contsum[nzeromask] / weights[nzeromask]
    expected = expected.reshape((outgridsize, outgridsize))
    expected = expected.T  # transposed for image plotting
    # expected[np.isnan(expected)] = 0.0  # convention in the slices

    # print(axis, shiftcenter, depth, periodic, weighted)
    # print("expected:\n", expected)
    # print("recovered:\n", img.v)
    assert_rel_equal(expected, img.v, 4)


# only axis-aligned; testing YTArbitraryGrid, YTCoveringGrid
@pytest.mark.parametrize("periodic", [True, False, (True, True, False)])
@pytest.mark.parametrize("wholebox", [True, False])
def test_sph_grid(
    periodic: bool | tuple[bool, bool, bool],
    wholebox: bool,
) -> None:
    bbox = np.array([[-1.0, 3.0], [1.0, 5.2], [-1.0, 3.0]])
    ds = fake_random_sph_ds(50, bbox, periodic=periodic)

    if not hasattr(periodic, "__len__"):
        periodic = (periodic,) * 3

    if wholebox:
        left = bbox[:, 0].copy()
        level = 2
        ncells = np.array([2**level] * 3)
        # print("left: ", left)
        # print("ncells: ", ncells)
        resgrid = ds.covering_grid(level, tuple(left), ncells)
        right = bbox[:, 1].copy()
        xedges = np.linspace(left[0], right[0], ncells[0] + 1)
        yedges = np.linspace(left[1], right[1], ncells[1] + 1)
        zedges = np.linspace(left[2], right[2], ncells[2] + 1)
    else:
        left = np.array([-1.0, 1.8, -1.0])
        right = np.array([2.5, 5.2, 2.5])
        ncells = np.array([3, 4, 4])
        resgrid = ds.arbitrary_grid(left, right, dims=ncells)
        xedges = np.linspace(left[0], right[0], ncells[0] + 1)
        yedges = np.linspace(left[1], right[1], ncells[1] + 1)
        zedges = np.linspace(left[2], right[2], ncells[2] + 1)
    res = resgrid["gas", "density"]
    xcens = 0.5 * (xedges[:-1] + xedges[1:])
    ycens = 0.5 * (yedges[:-1] + yedges[1:])
    zcens = 0.5 * (zedges[:-1] + zedges[1:])

    ad = ds.all_data()
    sphcoords = np.array(
        [
            (ad[("gas", "x")]).to("cm"),
            (ad[("gas", "y")]).to("cm"),
            (ad[("gas", "z")]).to("cm"),
        ]
    ).T
    gridx, gridy, gridz = np.meshgrid(xcens, ycens, zcens, indexing="ij")
    outshape = gridx.shape
    gridx = gridx.flatten()
    gridy = gridy.flatten()
    gridz = gridz.flatten()
    gridcoords = np.array([gridx, gridy, gridz]).T
    periods = bbox[:, 1] - bbox[:, 0]
    dists = distancematrix(gridcoords, sphcoords, periodic=periodic, periods=periods)
    sml = (ad[("gas", "smoothing_length")]).to("cm")
    normkern = cubicspline_python(dists / sml.v[np.newaxis, :])
    sphcontr = normkern / sml[np.newaxis, :] ** 3 * ad[("gas", "mass")]
    contsum = np.sum(sphcontr, axis=1)
    sphweights = (
        normkern
        / sml[np.newaxis, :] ** 3
        * ad[("gas", "mass")]
        / ad[("gas", "density")]
    )
    weights = np.sum(sphweights, axis=1)
    nzeromask = np.logical_not(weights == 0)
    expected = np.zeros(weights.shape, weights.dtype)
    expected[nzeromask] = contsum[nzeromask] / weights[nzeromask]
    expected = expected.reshape(outshape)
    # expected[np.isnan(expected)] = 0.0  # convention in the slices

    # print(axis, shiftcenter, depth, periodic, weighted)
    # print("expected:\n", expected)
    # print("recovered:\n", res.v)
    assert_rel_equal(expected, res.v, 4)



import yt

# Load the dataset
ds = yt.load("SecondOrderTris/RZ_p_no_parts_do_nothing_bcs_cone_out.e", step=-1)

# Create a line plot of the variables 'u' and 'v' with 1000 sampling points evenly
# spaced between the coordinates (0, 0, 0) and (0, 1, 0)
plot = yt.LinePlot(
    ds, [("all", "v"), ("all", "u")], (0.0, 0.0, 0.0), (0.0, 1.0, 0.0), 1000
)

# Add a legend
plot.annotate_legend(("all", "v"))

# Save the line plot
plot.save()



from numpy.testing import assert_equal

from yt.frontends.chombo.api import ChomboDataset, Orion2Dataset, PlutoDataset
from yt.testing import requires_file, requires_module, units_override_check
from yt.utilities.answer_testing.framework import (
    data_dir_load,
    requires_ds,
    small_patch_amr,
)

_fields = (
    ("gas", "density"),
    ("gas", "velocity_magnitude"),
    ("gas", "magnetic_field_x"),
)

gc = "GaussianCloud/data.0077.3d.hdf5"


@requires_ds(gc)
def test_gc():
    ds = data_dir_load(gc)
    assert_equal(str(ds), "data.0077.3d.hdf5")
    for test in small_patch_amr(ds, _fields):
        test_gc.__name__ = test.description
        yield test


tb = "TurbBoxLowRes/data.0005.3d.hdf5"


@requires_module("h5py")
@requires_ds(tb)
def test_tb():
    ds = data_dir_load(tb)
    assert_equal(str(ds), "data.0005.3d.hdf5")
    for test in small_patch_amr(ds, _fields):
        test_tb.__name__ = test.description
        yield test


iso = "IsothermalSphere/data.0000.3d.hdf5"


@requires_module("h5py")
@requires_ds(iso)
def test_iso():
    ds = data_dir_load(iso)
    assert_equal(str(ds), "data.0000.3d.hdf5")
    for test in small_patch_amr(ds, _fields):
        test_iso.__name__ = test.description
        yield test


_zp_fields = (("chombo", "rhs"), ("chombo", "phi"))
zp = "ZeldovichPancake/plt32.2d.hdf5"


@requires_module("h5py")
@requires_ds(zp)
def test_zp():
    ds = data_dir_load(zp)
    assert_equal(str(ds), "plt32.2d.hdf5")
    for test in small_patch_amr(ds, _zp_fields, input_center="c", input_weight="rhs"):
        test_zp.__name__ = test.description
        yield test


kho = "KelvinHelmholtz/data.0004.hdf5"


@requires_module("h5py")
@requires_ds(kho)
def test_kho():
    ds = data_dir_load(kho)
    assert_equal(str(ds), "data.0004.hdf5")
    for test in small_patch_amr(ds, _fields):
        test_kho.__name__ = test.description
        yield test


@requires_module("h5py")
@requires_file(zp)
def test_ChomboDataset():
    assert isinstance(data_dir_load(zp), ChomboDataset)


@requires_module("h5py")
@requires_file(gc)
def test_Orion2Dataset():
    assert isinstance(data_dir_load(gc), Orion2Dataset)


@requires_module("h5py")
@requires_file(kho)
def test_PlutoDataset():
    assert isinstance(data_dir_load(kho), PlutoDataset)


@requires_module("h5py")
@requires_file(zp)
def test_units_override_zp():
    units_override_check(zp)


@requires_module("h5py")
@requires_file(gc)
def test_units_override_gc():
    units_override_check(gc)


@requires_module("h5py")
@requires_file(kho)
def test_units_override_kho():
    units_override_check(kho)



import yt

ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
slc = yt.SlicePlot(ds, "x", ("gas", "density"))
slc.save("default_sliceplot.png")

slc.hide_axes()
slc.save("no_axes_sliceplot.png")

slc.hide_colorbar()
slc.save("no_axes_no_colorbar_sliceplot.png")

slc.show_axes()
slc.save("no_colorbar_sliceplot.png")



import contextlib

from more_itertools import always_iterable

from yt.data_objects.data_containers import YTFieldData
from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
)
from yt.utilities.exceptions import (
    YTDataSelectorNotImplemented,
    YTNonIndexedDataContainer,
)


def _non_indexed(name):
    def _func_non_indexed(self, *args, **kwargs):
        raise YTNonIndexedDataContainer(self)

    return _func_non_indexed


class ParticleContainer(YTSelectionContainer):
    _spatial = False
    _type_name = "particle_container"
    _skip_add = True
    _con_args = ("base_region", "base_selector", "data_files", "overlap_files")

    def __init__(
        self, base_region, base_selector, data_files, overlap_files=None, domain_id=-1
    ):
        if overlap_files is None:
            overlap_files = []
        self.field_data = YTFieldData()
        self.field_parameters = {}
        self.data_files = list(always_iterable(data_files))
        self.overlap_files = list(always_iterable(overlap_files))
        self.ds = self.data_files[0].ds
        self._last_mask = None
        self._last_selector_id = None
        self._current_particle_type = "all"
        # self._current_fluid_type = self.ds.default_fluid_type

        self.base_region = base_region
        self.base_selector = base_selector
        self._octree = None
        self._temp_spatial = False
        if isinstance(base_region, ParticleContainer):
            self._temp_spatial = base_region._temp_spatial
            self._octree = base_region._octree
        # To ensure there are not domains if global octree not used
        self.domain_id = -1

    def __reduce__(self):
        # we need to override the __reduce__ from data_containers as this method is not
        # a registered dataset method (i.e., ds.particle_container does not exist)
        arg_tuple = tuple(getattr(self, attr) for attr in self._con_args)
        return (self.__class__, arg_tuple)

    @property
    def selector(self):
        raise YTDataSelectorNotImplemented(self.oc_type_name)

    def select_particles(self, selector, x, y, z):
        mask = selector.select_points(x, y, z)
        return mask

    @contextlib.contextmanager
    def _expand_data_files(self):
        old_data_files = self.data_files
        old_overlap_files = self.overlap_files
        self.data_files = list(set(self.data_files + self.overlap_files))
        self.data_files.sort()
        self.overlap_files = []
        yield self
        self.data_files = old_data_files
        self.overlap_files = old_overlap_files

    def retrieve_ghost_zones(self, ngz, coarse_ghosts=False):
        gz_oct = self.octree.retrieve_ghost_zones(ngz, coarse_ghosts=coarse_ghosts)
        gz = ParticleContainer(
            gz_oct.base_region,
            gz_oct.base_selector,
            gz_oct.data_files,
            overlap_files=gz_oct.overlap_files,
            selector_mask=gz_oct.selector_mask,
            domain_id=gz_oct.domain_id,
        )
        gz._octree = gz_oct
        return gz

    select_blocks = _non_indexed("select_blocks")
    deposit = _non_indexed("deposit")
    smooth = _non_indexed("smooth")
    select_icoords = _non_indexed("select_icoords")
    select_fcoords = _non_indexed("select_fcoords")
    select_fwidth = _non_indexed("select_fwidth")
    select_ires = _non_indexed("select_ires")
    select = _non_indexed("select")
    count = _non_indexed("count")
    count_particles = _non_indexed("count_particles")



import os
import time
import uuid
import warnings
import weakref
from collections import defaultdict
from functools import cached_property

import numpy as np
from more_itertools import always_iterable

from yt.config import ytcfg
from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.fields.field_info_container import FieldInfoContainer
from yt.funcs import mylog, setdefaultattr
from yt.geometry.api import Geometry
from yt.geometry.geometry_handler import YTDataChunk
from yt.geometry.grid_geometry_handler import GridIndex
from yt.units import dimensions
from yt.units.unit_lookup_table import (  # type: ignore
    default_unit_symbol_lut,
    unit_prefixes,
)
from yt.units.unit_object import UnitParseError  # type: ignore
from yt.units.yt_array import YTQuantity
from yt.utilities.decompose import decompose_array, get_psize
from yt.utilities.file_handler import FITSFileHandler
from yt.utilities.io_handler import io_registry
from yt.utilities.on_demand_imports import NotAModule, _astropy

from .fields import FITSFieldInfo, WCSFITSFieldInfo, YTFITSFieldInfo

lon_prefixes = ["X", "RA", "GLON", "LINEAR"]
lat_prefixes = ["Y", "DEC", "GLAT", "LINEAR"]

spec_names = {"V": "Velocity", "F": "Frequency", "E": "Energy", "W": "Wavelength"}

space_prefixes = list(set(lon_prefixes + lat_prefixes))
unique_sky_prefixes = set(space_prefixes)
unique_sky_prefixes.difference_update({"X", "Y", "LINEAR"})
sky_prefixes = list(unique_sky_prefixes)
spec_prefixes = list(spec_names.keys())


class FITSGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, id, index, level):
        AMRGridPatch.__init__(self, id, filename=index.index_filename, index=index)
        self.Parent = None
        self.Children = []
        self.Level = 0


class FITSHierarchy(GridIndex):
    grid = FITSGrid

    def __init__(self, ds, dataset_type="fits"):
        self.dataset_type = dataset_type
        self.field_indexes = {}
        self.dataset = weakref.proxy(ds)
        # for now, the index file is the dataset
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.path.dirname(self.index_filename)
        self._handle = ds._handle
        self.float_type = np.float64
        GridIndex.__init__(self, ds, dataset_type)

    def _initialize_data_storage(self):
        pass

    def _guess_name_from_units(self, units):
        field_from_unit = {"Jy": "intensity", "K": "temperature"}
        for k, v in field_from_unit.items():
            if k in units:
                mylog.warning(
                    "Guessing this is a %s field based on its units of %s.", v, k
                )
                return v
        return None

    def _determine_image_units(self, bunit):
        try:
            try:
                # First let AstroPy attempt to figure the unit out
                u = 1.0 * _astropy.units.Unit(bunit, format="fits")
                u = YTQuantity.from_astropy(u).units
            except ValueError:
                try:
                    # Let yt try it by itself
                    u = self.ds.quan(1.0, bunit).units
                except UnitParseError:
                    return "dimensionless"
            return str(u)
        except KeyError:
            return "dimensionless"

    def _ensure_same_dims(self, hdu):
        ds = self.dataset
        conditions = [hdu.header["naxis"] != ds.primary_header["naxis"]]
        for i in range(ds.naxis):
            nax = "naxis%d" % (i + 1)
            conditions.append(hdu.header[nax] != ds.primary_header[nax])
        if np.any(conditions):
            return False
        else:
            return True

    def _detect_output_fields(self):
        self.field_list = []
        self._axis_map = {}
        self._file_map = {}
        self._ext_map = {}
        self._scale_map = {}
        dup_field_index = {}
        # Since FITS header keywords are case-insensitive, we only pick a subset of
        # prefixes, ones that we expect to end up in headers.
        known_units = {unit.lower(): unit for unit in self.ds.unit_registry.lut}
        for unit in list(known_units.values()):
            if unit in self.ds.unit_registry.prefixable_units:
                for p in ["n", "u", "m", "c", "k"]:
                    known_units[(p + unit).lower()] = p + unit
        # We create a field from each slice on the 4th axis
        if self.dataset.naxis == 4:
            naxis4 = self.dataset.primary_header["naxis4"]
        else:
            naxis4 = 1
        for i, fits_file in enumerate(self.dataset._handle._fits_files):
            for j, hdu in enumerate(fits_file):
                if (
                    isinstance(hdu, _astropy.pyfits.BinTableHDU)
                    or hdu.header["naxis"] == 0
                ):
                    continue
                if self._ensure_same_dims(hdu):
                    units = self._determine_image_units(hdu.header["bunit"])
                    try:
                        # Grab field name from btype
                        fname = hdu.header["btype"]
                    except KeyError:
                        # Try to guess the name from the units
                        fname = self._guess_name_from_units(units)
                        # When all else fails
                        if fname is None:
                            fname = "image_%d" % (j)
                    if self.ds.num_files > 1 and fname.startswith("image"):
                        fname += "_file_%d" % (i)
                    if ("fits", fname) in self.field_list:
                        if fname in dup_field_index:
                            dup_field_index[fname] += 1
                        else:
                            dup_field_index[fname] = 1
                        mylog.warning(
                            "This field has the same name as a previously loaded "
                            "field. Changing the name from %s to %s_%d. To avoid "
                            "this, change one of the BTYPE header keywords.",
                            fname,
                            fname,
                            dup_field_index[fname],
                        )
                        fname += "_%d" % (dup_field_index[fname])
                    for k in range(naxis4):
                        if naxis4 > 1:
                            fname += "_%s_%d" % (hdu.header["CTYPE4"], k + 1)
                        self._axis_map[fname] = k
                        self._file_map[fname] = fits_file
                        self._ext_map[fname] = j
                        self._scale_map[fname] = [0.0, 1.0]
                        if "bzero" in hdu.header:
                            self._scale_map[fname][0] = hdu.header["bzero"]
                        if "bscale" in hdu.header:
                            self._scale_map[fname][1] = hdu.header["bscale"]
                        self.field_list.append(("fits", fname))
                        self.dataset.field_units[fname] = units
                        mylog.info("Adding field %s to the list of fields.", fname)
                        if units == "dimensionless":
                            mylog.warning(
                                "Could not determine dimensions for field %s, "
                                "setting to dimensionless.",
                                fname,
                            )
                else:
                    mylog.warning(
                        "Image block %s does not have the same dimensions "
                        "as the primary and will not be available as a field.",
                        hdu.name.lower(),
                    )

    def _count_grids(self):
        self.num_grids = self.ds.parameters["nprocs"]

    def _parse_index(self):
        ds = self.dataset

        # If nprocs > 1, decompose the domain into virtual grids
        if self.num_grids > 1:
            self._domain_decomp()
        else:
            self.grid_left_edge[0, :] = ds.domain_left_edge
            self.grid_right_edge[0, :] = ds.domain_right_edge
            self.grid_dimensions[0] = ds.domain_dimensions

        self.grid_levels.flat[:] = 0
        self.grids = np.empty(self.num_grids, dtype="object")
        for i in range(self.num_grids):
            self.grids[i] = self.grid(i, self, self.grid_levels[i, 0])

    def _domain_decomp(self):
        bbox = np.array(
            [self.ds.domain_left_edge, self.ds.domain_right_edge]
        ).transpose()
        dims = self.ds.domain_dimensions
        psize = get_psize(dims, self.num_grids)
        gle, gre, shapes, slices, _ = decompose_array(dims, psize, bbox)
        self.grid_left_edge = self.ds.arr(gle, "code_length")
        self.grid_right_edge = self.ds.arr(gre, "code_length")
        self.grid_dimensions = np.array(shapes, dtype="int32")

    def _populate_grid_objects(self):
        for i in range(self.num_grids):
            self.grids[i]._prepare_grid()
            self.grids[i]._setup_dx()
        self.max_level = 0

    def _setup_derived_fields(self):
        super()._setup_derived_fields()
        [self.dataset.conversion_factors[field] for field in self.field_list]
        for field in self.field_list:
            if field not in self.derived_field_list:
                self.derived_field_list.append(field)

        for field in self.derived_field_list:
            f = self.dataset.field_info[field]
            if f.is_alias:
                # Translating an already-converted field
                self.dataset.conversion_factors[field] = 1.0

    def _setup_data_io(self):
        self.io = io_registry[self.dataset_type](self.dataset)

    def _chunk_io(self, dobj, cache=True, local_only=False):
        # local_only is only useful for inline datasets and requires
        # implementation by subclasses.
        gfiles = defaultdict(list)
        gobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for g in gobjs:
            gfiles[g.id].append(g)
        for fn in sorted(gfiles):
            gs = gfiles[fn]
            yield YTDataChunk(
                dobj, "io", gs, self._count_selection(dobj, gs), cache=cache
            )


def find_primary_header(fileh):
    # Sometimes the primary hdu doesn't have an image
    if len(fileh) > 1 and fileh[0].header["naxis"] == 0:
        first_image = 1
    else:
        first_image = 0
    header = fileh[first_image].header
    return header, first_image


def check_fits_valid(filename):
    ext = filename.rsplit(".", 1)[-1]
    if ext.upper() in ("GZ", "FZ"):
        # We don't know for sure that there will be > 1
        ext = filename.rsplit(".", 1)[0].rsplit(".", 1)[-1]
    if ext.upper() not in ("FITS", "FTS"):
        return None
    elif isinstance(_astropy.pyfits, NotAModule):
        raise RuntimeError(
            "This appears to be a FITS file, but AstroPy is not installed."
        )
    try:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning, append=True)
            fileh = _astropy.pyfits.open(filename)
        header, _ = find_primary_header(fileh)
        if header["naxis"] >= 2:
            return fileh
        else:
            fileh.close()
    except Exception:
        pass
    return None


def check_sky_coords(filename, ndim):
    fileh = check_fits_valid(filename)
    if fileh is not None:
        try:
            if len(fileh) > 1 and fileh[1].name == "EVENTS" and ndim == 2:
                fileh.close()
                return True
            else:
                header, _ = find_primary_header(fileh)
                if header["naxis"] < ndim:
                    return False
                axis_names = [
                    header.get("ctype%d" % (i + 1), "") for i in range(header["naxis"])
                ]
                if len(axis_names) == 3 and axis_names.count("LINEAR") == 2:
                    return any(a[0] in spec_prefixes for a in axis_names)
                x = find_axes(axis_names, sky_prefixes + spec_prefixes)
                fileh.close()
                return x >= ndim
        except Exception:
            pass
    return False


class FITSDataset(Dataset):
    _load_requirements = ["astropy"]
    _index_class = FITSHierarchy
    _field_info_class: type[FieldInfoContainer] = FITSFieldInfo
    _dataset_type = "fits"
    _handle = None

    def __init__(
        self,
        filename,
        dataset_type="fits",
        auxiliary_files=None,
        nprocs=None,
        storage_filename=None,
        nan_mask=None,
        suppress_astropy_warnings=True,
        parameters=None,
        units_override=None,
        unit_system="cgs",
    ):
        if parameters is None:
            parameters = {}
        parameters["nprocs"] = nprocs
        self.specified_parameters = parameters

        if suppress_astropy_warnings:
            warnings.filterwarnings("ignore", module="astropy", append=True)

        self.filenames = [filename] + list(always_iterable(auxiliary_files))
        self.num_files = len(self.filenames)
        self.fluid_types += ("fits",)
        if nan_mask is None:
            self.nan_mask = {}
        elif isinstance(nan_mask, float):
            self.nan_mask = {"all": nan_mask}
        elif isinstance(nan_mask, dict):
            self.nan_mask = nan_mask
        self._handle = FITSFileHandler(self.filenames[0])
        if isinstance(
            self.filenames[0], _astropy.pyfits.hdu.image._ImageBaseHDU
        ) or isinstance(self.filenames[0], _astropy.pyfits.HDUList):
            fn = f"InMemoryFITSFile_{uuid.uuid4().hex}"
        else:
            fn = self.filenames[0]
        self._handle._fits_files.append(self._handle)
        if self.num_files > 1:
            for fits_file in auxiliary_files:
                if isinstance(fits_file, _astropy.pyfits.hdu.image._ImageBaseHDU):
                    f = _astropy.pyfits.HDUList([fits_file])
                elif isinstance(fits_file, _astropy.pyfits.HDUList):
                    f = fits_file
                else:
                    if os.path.exists(fits_file):
                        fn = fits_file
                    else:
                        fn = os.path.join(ytcfg.get("yt", "test_data_dir"), fits_file)
                    f = _astropy.pyfits.open(
                        fn, memmap=True, do_not_scale_image_data=True, ignore_blank=True
                    )
                self._handle._fits_files.append(f)

        self.refine_by = 2

        Dataset.__init__(
            self,
            fn,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )
        self.storage_filename = storage_filename

    def _set_code_unit_attributes(self):
        """
        Generates the conversion to various physical _units based on the
        parameter file
        """
        if getattr(self, "length_unit", None) is None:
            default_length_units = [
                u for u, v in default_unit_symbol_lut.items() if str(v[1]) == "(length)"
            ]
            more_length_units = []
            for unit in default_length_units:
                if unit in self.unit_registry.prefixable_units:
                    more_length_units += [prefix + unit for prefix in unit_prefixes]
            default_length_units += more_length_units
            file_units = []
            cunits = [self.wcs.wcs.cunit[i] for i in range(self.dimensionality)]
            for unit in (_.to_string() for _ in cunits):
                if unit in default_length_units:
                    file_units.append(unit)
            if len(set(file_units)) == 1:
                length_factor = self.wcs.wcs.cdelt[0]
                length_unit = str(file_units[0])
                mylog.info("Found length units of %s.", length_unit)
            else:
                self.no_cgs_equiv_length = True
                mylog.warning("No length conversion provided. Assuming 1 = 1 cm.")
                length_factor = 1.0
                length_unit = "cm"
            setdefaultattr(self, "length_unit", self.quan(length_factor, length_unit))
        for unit, cgs in [("time", "s"), ("mass", "g")]:
            # We set these to cgs for now, but they may have been overridden
            if getattr(self, unit + "_unit", None) is not None:
                continue
            mylog.warning("Assuming 1.0 = 1.0 %s", cgs)
            setdefaultattr(self, f"{unit}_unit", self.quan(1.0, cgs))
        self.magnetic_unit = np.sqrt(
            4 * np.pi * self.mass_unit / (self.time_unit**2 * self.length_unit)
        )
        self.magnetic_unit.convert_to_units("gauss")
        self.velocity_unit = self.length_unit / self.time_unit

    @property
    def filename(self) -> str:
        if self._input_filename.startswith("InMemory"):
            return self._input_filename
        else:
            return super().filename

    @cached_property
    def unique_identifier(self) -> str:
        if self.filename.startswith("InMemory"):
            return str(time.time())
        else:
            return super().unique_identifier

    def _parse_parameter_file(self):
        self._determine_structure()
        self._determine_axes()

        # Determine dimensionality

        self.dimensionality = self.naxis
        self.geometry = Geometry.CARTESIAN

        # Sometimes a FITS file has a 4D datacube, in which case
        # we take the 4th axis and assume it consists of different fields.
        if self.dimensionality == 4:
            self.dimensionality = 3

        self._determine_wcs()

        self.current_time = 0.0

        self.domain_dimensions = np.array(self.dims)[: self.dimensionality]
        if self.dimensionality == 2:
            self.domain_dimensions = np.append(self.domain_dimensions, [1])
        self._determine_bbox()

        # Get the simulation time
        try:
            self.current_time = self.parameters["time"]
        except Exception:
            mylog.warning("Cannot find time")
            self.current_time = 0.0
            pass

        # For now we'll ignore these
        self._periodicity = (False,) * 3
        self.current_redshift = 0.0
        self.omega_lambda = 0.0
        self.omega_matter = 0.0
        self.hubble_constant = 0.0
        self.cosmological_simulation = 0

        self._determine_nprocs()

        # Now we can set up some of our parameters for convenience.
        for k, v in self.primary_header.items():
            self.parameters[k] = v
        # Remove potential default keys
        self.parameters.pop("", None)

    def _determine_nprocs(self):
        # If nprocs is None, do some automatic decomposition of the domain
        if self.specified_parameters["nprocs"] is None:
            nprocs = np.around(
                np.prod(self.domain_dimensions) / 32**self.dimensionality
            ).astype("int64")
            self.parameters["nprocs"] = max(min(nprocs, 512), 1)
        else:
            self.parameters["nprocs"] = self.specified_parameters["nprocs"]

    def _determine_structure(self):
        self.primary_header, self.first_image = find_primary_header(self._handle)
        self.naxis = self.primary_header["naxis"]
        self.axis_names = [
            self.primary_header.get("ctype%d" % (i + 1), "LINEAR")
            for i in range(self.naxis)
        ]
        self.dims = [
            self.primary_header["naxis%d" % (i + 1)] for i in range(self.naxis)
        ]

    def _determine_wcs(self):
        wcs = _astropy.pywcs.WCS(header=self.primary_header)
        if self.naxis == 4:
            self.wcs = _astropy.pywcs.WCS(naxis=3)
            self.wcs.wcs.crpix = wcs.wcs.crpix[:3]
            self.wcs.wcs.cdelt = wcs.wcs.cdelt[:3]
            self.wcs.wcs.crval = wcs.wcs.crval[:3]
            self.wcs.wcs.cunit = [str(unit) for unit in wcs.wcs.cunit][:3]
            self.wcs.wcs.ctype = list(wcs.wcs.ctype)[:3]
        else:
            self.wcs = wcs

    def _determine_bbox(self):
        domain_left_edge = np.array([0.5] * 3)
        domain_right_edge = np.array(
            [float(dim) + 0.5 for dim in self.domain_dimensions]
        )

        if self.dimensionality == 2:
            domain_left_edge[-1] = 0.5
            domain_right_edge[-1] = 1.5

        self.domain_left_edge = domain_left_edge
        self.domain_right_edge = domain_right_edge

    def _determine_axes(self):
        self.lat_axis = 1
        self.lon_axis = 0
        self.lat_name = "Y"
        self.lon_name = "X"

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False

        try:
            fileh = check_fits_valid(filename)
        except Exception:
            return False

        if fileh is None:
            return False
        else:
            fileh.close()
            return True

    @classmethod
    def _guess_candidates(cls, base, directories, files):
        candidates = []
        for fn, fnl in ((_, _.lower()) for _ in files):
            if (
                fnl.endswith(".fits")
                or fnl.endswith(".fits.gz")
                or fnl.endswith(".fits.fz")
            ):
                candidates.append(fn)
        # FITS files don't preclude subdirectories
        return candidates, True

    def close(self):
        self._handle.close()


def find_axes(axis_names, prefixes):
    x = 0
    for p in prefixes:
        y = np.char.startswith(axis_names, p)
        x += np.any(y)
    return x


class YTFITSDataset(FITSDataset):
    _load_requirements = ["astropy"]
    _field_info_class = YTFITSFieldInfo

    def _parse_parameter_file(self):
        super()._parse_parameter_file()
        # Get the current time
        if "time" in self.primary_header:
            self.current_time = self.primary_header["time"]

    def _set_code_unit_attributes(self):
        """
        Generates the conversion to various physical _units based on the parameter file
        """
        for unit, cgs in [
            ("length", "cm"),
            ("time", "s"),
            ("mass", "g"),
            ("velocity", "cm/s"),
            ("magnetic", "gauss"),
        ]:
            if unit == "magnetic":
                short_unit = "bfunit"
            else:
                short_unit = f"{unit[0]}unit"
            if short_unit in self.primary_header:
                # units should now be in header
                u = self.quan(
                    self.primary_header[short_unit],
                    self.primary_header.comments[short_unit].strip("[]"),
                )
                mylog.info("Found %s units of %s.", unit, u)
            else:
                if unit == "length":
                    # Falling back to old way of getting units for length
                    # in old files
                    u = self.quan(1.0, str(self.wcs.wcs.cunit[0]))
                    mylog.info("Found %s units of %s.", unit, u)
                else:
                    # Give up otherwise
                    u = self.quan(1.0, cgs)
                    mylog.warning(
                        "No unit for %s found. Assuming 1.0 code_%s = 1.0 %s",
                        unit,
                        unit,
                        cgs,
                    )
            setdefaultattr(self, f"{unit}_unit", u)

    def _determine_bbox(self):
        dx = np.zeros(3)
        dx[: self.dimensionality] = self.wcs.wcs.cdelt
        domain_left_edge = np.zeros(3)
        domain_left_edge[: self.dimensionality] = self.wcs.wcs.crval - dx[
            : self.dimensionality
        ] * (self.wcs.wcs.crpix - 0.5)
        domain_right_edge = domain_left_edge + dx * self.domain_dimensions

        if self.dimensionality == 2:
            domain_left_edge[-1] = 0.0
            domain_right_edge[-1] = dx[0]

        self.domain_left_edge = domain_left_edge
        self.domain_right_edge = domain_right_edge

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False

        try:
            fileh = check_fits_valid(filename)
        except Exception:
            return False

        if fileh is None:
            return False
        else:
            if "WCSNAME" in fileh[0].header:
                isyt = fileh[0].header["WCSNAME"].strip() == "yt"
            else:
                isyt = False
            fileh.close()
            return isyt


class SkyDataFITSDataset(FITSDataset):
    _load_requirements = ["astropy"]
    _field_info_class = WCSFITSFieldInfo

    def _determine_wcs(self):
        super()._determine_wcs()
        end = min(self.dimensionality + 1, 4)
        self.ctypes = np.array(
            [self.primary_header["CTYPE%d" % (i)] for i in range(1, end)]
        )
        self.wcs_2d = self.wcs

    def _parse_parameter_file(self):
        super()._parse_parameter_file()

        end = min(self.dimensionality + 1, 4)

        self.geometry = Geometry.SPECTRAL_CUBE

        log_str = "Detected these axes: " + "%s " * len(self.ctypes)
        mylog.info(log_str, *self.ctypes)

        self.lat_axis = np.zeros((end - 1), dtype="bool")
        for p in lat_prefixes:
            self.lat_axis += np.char.startswith(self.ctypes, p)
        self.lat_axis = np.where(self.lat_axis)[0][0]
        self.lat_name = self.ctypes[self.lat_axis].split("-")[0].lower()

        self.lon_axis = np.zeros((end - 1), dtype="bool")
        for p in lon_prefixes:
            self.lon_axis += np.char.startswith(self.ctypes, p)
        self.lon_axis = np.where(self.lon_axis)[0][0]
        self.lon_name = self.ctypes[self.lon_axis].split("-")[0].lower()

        if self.lat_axis == self.lon_axis and self.lat_name == self.lon_name:
            self.lat_axis = 1
            self.lon_axis = 0
            self.lat_name = "Y"
            self.lon_name = "X"

        self.spec_axis = 2
        self.spec_name = "z"
        self.spec_unit = ""

    def _set_code_unit_attributes(self):
        super()._set_code_unit_attributes()
        units = self.wcs_2d.wcs.cunit[0]
        if units == "deg":
            units = "degree"
        if units == "rad":
            units = "radian"
        pixel_area = np.prod(np.abs(self.wcs_2d.wcs.cdelt))
        pixel_area = self.quan(pixel_area, f"{units}**2").in_cgs()
        pixel_dims = pixel_area.units.dimensions
        self.unit_registry.add("pixel", float(pixel_area.value), dimensions=pixel_dims)
        if "beam_size" in self.specified_parameters:
            beam_size = self.specified_parameters["beam_size"]
            beam_size = self.quan(beam_size[0], beam_size[1]).in_cgs().value
            self.unit_registry.add("beam", beam_size, dimensions=dimensions.solid_angle)

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False

        try:
            return check_sky_coords(filename, ndim=2)
        except Exception:
            return False


class SpectralCubeFITSHierarchy(FITSHierarchy):
    def _domain_decomp(self):
        dz = self.ds.quan(1.0, "code_length") * self.ds.spectral_factor
        self.grid_dimensions[:, 2] = np.around(
            float(self.ds.domain_dimensions[2]) / self.num_grids
        ).astype("int64")
        self.grid_dimensions[-1, 2] += self.ds.domain_dimensions[2] % self.num_grids
        self.grid_left_edge[0, 2] = self.ds.domain_left_edge[2]
        self.grid_left_edge[1:, 2] = (
            self.ds.domain_left_edge[2] + np.cumsum(self.grid_dimensions[:-1, 2]) * dz
        )
        self.grid_right_edge[:, 2] = (
            self.grid_left_edge[:, 2] + self.grid_dimensions[:, 2] * dz
        )
        self.grid_left_edge[:, :2] = self.ds.domain_left_edge[:2]
        self.grid_right_edge[:, :2] = self.ds.domain_right_edge[:2]
        self.grid_dimensions[:, :2] = self.ds.domain_dimensions[:2]


class SpectralCubeFITSDataset(SkyDataFITSDataset):
    _load_requirements = ["astropy"]
    _index_class = SpectralCubeFITSHierarchy

    def __init__(
        self,
        filename,
        auxiliary_files=None,
        nprocs=None,
        storage_filename=None,
        nan_mask=None,
        spectral_factor=1.0,
        suppress_astropy_warnings=True,
        parameters=None,
        units_override=None,
        unit_system="cgs",
    ):
        if auxiliary_files is None:
            auxiliary_files = []
        self.spectral_factor = spectral_factor
        super().__init__(
            filename,
            nprocs=nprocs,
            auxiliary_files=auxiliary_files,
            storage_filename=storage_filename,
            suppress_astropy_warnings=suppress_astropy_warnings,
            nan_mask=nan_mask,
            parameters=parameters,
            units_override=units_override,
            unit_system=unit_system,
        )

    def _parse_parameter_file(self):
        super()._parse_parameter_file()

        self.geometry = Geometry.SPECTRAL_CUBE

        end = min(self.dimensionality + 1, 4)

        self.spec_axis = np.zeros(end - 1, dtype="bool")
        for p in spec_names.keys():
            self.spec_axis += np.char.startswith(self.ctypes, p)
        self.spec_axis = np.where(self.spec_axis)[0][0]
        self.spec_name = spec_names[self.ctypes[self.spec_axis].split("-")[0][0]]

        # Extract a subimage from a WCS object
        self.wcs_2d = self.wcs.sub(["longitude", "latitude"])

        self._p0 = self.wcs.wcs.crpix[self.spec_axis]
        self._dz = self.wcs.wcs.cdelt[self.spec_axis]
        self._z0 = self.wcs.wcs.crval[self.spec_axis]
        self.spec_unit = str(self.wcs.wcs.cunit[self.spec_axis])

        if self.spectral_factor == "auto":
            self.spectral_factor = float(
                max(self.domain_dimensions[[self.lon_axis, self.lat_axis]])
            )
            self.spectral_factor /= self.domain_dimensions[self.spec_axis]
            mylog.info("Setting the spectral factor to %f", self.spectral_factor)
        Dz = (
            self.domain_right_edge[self.spec_axis]
            - self.domain_left_edge[self.spec_axis]
        )
        dre = self.domain_right_edge.copy()
        dre[self.spec_axis] = (
            self.domain_left_edge[self.spec_axis] + self.spectral_factor * Dz
        )
        self.domain_right_edge = dre
        self._dz /= self.spectral_factor
        self._p0 = (self._p0 - 0.5) * self.spectral_factor + 0.5

    def _determine_nprocs(self):
        # If nprocs is None, do some automatic decomposition of the domain
        if self.specified_parameters["nprocs"] is None:
            nprocs = np.around(self.domain_dimensions[2] / 8).astype("int64")
            self.parameters["nprocs"] = max(min(nprocs, 512), 1)
        else:
            self.parameters["nprocs"] = self.specified_parameters["nprocs"]

    def spec2pixel(self, spec_value):
        sv = self.arr(spec_value).in_units(self.spec_unit)
        return self.arr((sv.v - self._z0) / self._dz + self._p0, "code_length")

    def pixel2spec(self, pixel_value):
        pv = self.arr(pixel_value, "code_length")
        return self.arr((pv.v - self._p0) * self._dz + self._z0, self.spec_unit)

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False

        try:
            return check_sky_coords(filename, ndim=3)
        except Exception:
            return False


class EventsFITSHierarchy(FITSHierarchy):
    def _detect_output_fields(self):
        ds = self.dataset
        self.field_list = []
        for k, v in ds.events_info.items():
            fname = "event_" + k
            mylog.info("Adding field %s to the list of fields.", fname)
            self.field_list.append(("io", fname))
            if k in ["x", "y"]:
                field_unit = "code_length"
            else:
                field_unit = v
            self.dataset.field_units["io", fname] = field_unit
        return

    def _parse_index(self):
        super()._parse_index()
        try:
            self.grid_particle_count[:] = self.dataset.primary_header["naxis2"]
        except KeyError:
            self.grid_particle_count[:] = 0.0
        self._particle_indices = np.zeros(self.num_grids + 1, dtype="int64")
        self._particle_indices[1] = self.grid_particle_count.squeeze()


class EventsFITSDataset(SkyDataFITSDataset):
    _load_requirements = ["astropy"]
    _index_class = EventsFITSHierarchy

    def __init__(
        self,
        filename,
        storage_filename=None,
        suppress_astropy_warnings=True,
        reblock=1,
        parameters=None,
        units_override=None,
        unit_system="cgs",
    ):
        self.reblock = reblock
        super().__init__(
            filename,
            nprocs=1,
            storage_filename=storage_filename,
            parameters=parameters,
            suppress_astropy_warnings=suppress_astropy_warnings,
            units_override=units_override,
            unit_system=unit_system,
        )

    def _determine_structure(self):
        self.first_image = 1
        self.primary_header = self._handle[self.first_image].header
        self.naxis = 2

    def _determine_wcs(self):
        self.wcs = _astropy.pywcs.WCS(naxis=2)
        self.events_info = {}
        for k, v in self.primary_header.items():
            if k.startswith("TTYP"):
                if v.lower() in ["x", "y"]:
                    num = k.replace("TTYPE", "")
                    self.events_info[v.lower()] = (
                        self.primary_header["TLMIN" + num],
                        self.primary_header["TLMAX" + num],
                        self.primary_header["TCTYP" + num],
                        self.primary_header["TCRVL" + num],
                        self.primary_header["TCDLT" + num],
                        self.primary_header["TCRPX" + num],
                    )
                elif v.lower() in ["energy", "time"]:
                    num = k.replace("TTYPE", "")
                    unit = self.primary_header["TUNIT" + num].lower()
                    if unit.endswith("ev"):
                        unit = unit.replace("ev", "eV")
                    self.events_info[v.lower()] = unit
        self.axis_names = [self.events_info[ax][2] for ax in ["x", "y"]]
        self.wcs.wcs.cdelt = [
            self.events_info["x"][4] * self.reblock,
            self.events_info["y"][4] * self.reblock,
        ]
        self.wcs.wcs.crpix = [
            (self.events_info["x"][5] - 0.5) / self.reblock + 0.5,
            (self.events_info["y"][5] - 0.5) / self.reblock + 0.5,
        ]
        self.wcs.wcs.ctype = [self.events_info["x"][2], self.events_info["y"][2]]
        self.wcs.wcs.cunit = ["deg", "deg"]
        self.wcs.wcs.crval = [self.events_info["x"][3], self.events_info["y"][3]]
        self.dims = [
            (self.events_info["x"][1] - self.events_info["x"][0]) / self.reblock,
            (self.events_info["y"][1] - self.events_info["y"][0]) / self.reblock,
        ]
        self.ctypes = self.axis_names
        self.wcs_2d = self.wcs

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False

        try:
            fileh = check_fits_valid(filename)
        except Exception:
            return False
        if fileh is not None:
            try:
                valid = fileh[1].name == "EVENTS"
                fileh.close()
                return valid
            except Exception:
                pass
        return False



import abc

from yt._typing import FieldKey

domain_context_registry = {}


class DomainContext(abc.ABC):
    class __metaclass__(type):
        def __init__(cls, name, b, d):
            type.__init__(cls, name, b, d)
            domain_context_registry[name] = cls

    _known_fluid_fields: tuple[FieldKey, ...]
    _known_particle_fields: tuple[FieldKey, ...]

    def __init__(self, ds):
        self.ds = ds



import matplotlib.pyplot as plt

import yt

ds = yt.load("GasSloshing/sloshing_nomag2_hdf5_plt_cnt_0150")

# Get a sphere object

sp = ds.sphere(ds.domain_center, (500.0, "kpc"))

# Bin up the data from the sphere into a radial profile

rp = yt.create_profile(
    sp,
    "radius",
    [("gas", "density"), ("gas", "temperature")],
    units={"radius": "kpc"},
    logs={"radius": False},
)

# Make plots using matplotlib

fig = plt.figure()
ax = fig.add_subplot(111)

# Plot the density as a log-log plot using the default settings
dens_plot = ax.loglog(rp.x.value, rp["gas", "density"].value)

# Here we set the labels of the plot axes

ax.set_xlabel(r"$\mathrm{r\ (kpc)}$")
ax.set_ylabel(r"$\mathrm{\rho\ (g\ cm^{-3})}$")

# Save the default plot

fig.savefig("density_profile_default.png" % ds)

# The "dens_plot" object is a list of plot objects. In our case we only have one,
# so we index the list by '0' to get it.

# Plot using dashed red lines

dens_plot[0].set_linestyle("--")
dens_plot[0].set_color("red")

fig.savefig("density_profile_dashed_red.png")

# Increase the line width and add points in the shape of x's

dens_plot[0].set_linewidth(5)
dens_plot[0].set_marker("x")
dens_plot[0].set_markersize(10)

fig.savefig("density_profile_thick_with_xs.png")



import numpy as np
from numpy.testing import assert_allclose, assert_equal

from yt.frontends.boxlib.api import (
    AMReXDataset,
    CastroDataset,
    MaestroDataset,
    NyxDataset,
    OrionDataset,
    WarpXDataset,
)
from yt.loaders import load
from yt.testing import (
    disable_dataset_cache,
    requires_file,
    units_override_check,
)
from yt.utilities.answer_testing.framework import (
    GridValuesTest,
    data_dir_load,
    requires_ds,
    small_patch_amr,
)

# We don't do anything needing ghost zone generation right now, because these
# are non-periodic datasets.
_orion_fields = (
    ("gas", "temperature"),
    ("gas", "density"),
    ("gas", "velocity_magnitude"),
)
_nyx_fields = (
    ("boxlib", "Ne"),
    ("boxlib", "Temp"),
    ("boxlib", "particle_mass_density"),
)
_warpx_fields = (("mesh", "Ex"), ("mesh", "By"), ("mesh", "jz"))
_castro_fields = (
    ("boxlib", "Temp"),
    ("gas", "density"),
    ("boxlib", "particle_count"),
)

radadvect = "RadAdvect/plt00000"


@requires_ds(radadvect)
def test_radadvect():
    ds = data_dir_load(radadvect)
    assert_equal(str(ds), "plt00000")
    for test in small_patch_amr(ds, _orion_fields):
        test_radadvect.__name__ = test.description
        yield test


rt = "RadTube/plt00500"


@requires_ds(rt)
def test_radtube():
    ds = data_dir_load(rt)
    assert_equal(str(ds), "plt00500")
    for test in small_patch_amr(ds, _orion_fields):
        test_radtube.__name__ = test.description
        yield test


star = "StarParticles/plrd01000"


@requires_ds(star)
def test_star():
    ds = data_dir_load(star)
    assert_equal(str(ds), "plrd01000")
    for test in small_patch_amr(ds, _orion_fields):
        test_star.__name__ = test.description
        yield test


LyA = "Nyx_LyA/plt00000"


@requires_ds(LyA)
def test_LyA():
    ds = data_dir_load(LyA)
    assert_equal(str(ds), "plt00000")
    for test in small_patch_amr(
        ds, _nyx_fields, input_center="c", input_weight=("boxlib", "Ne")
    ):
        test_LyA.__name__ = test.description
        yield test


@requires_file(LyA)
def test_nyx_particle_io():
    ds = data_dir_load(LyA)

    grid = ds.index.grids[0]
    npart_grid_0 = 7908  # read directly from the header
    assert_equal(grid["all", "particle_position_x"].size, npart_grid_0)
    assert_equal(grid["DM", "particle_position_y"].size, npart_grid_0)
    assert_equal(grid["all", "particle_position_z"].size, npart_grid_0)

    ad = ds.all_data()
    npart = 32768  # read directly from the header
    assert_equal(ad["all", "particle_velocity_x"].size, npart)
    assert_equal(ad["DM", "particle_velocity_y"].size, npart)
    assert_equal(ad["all", "particle_velocity_z"].size, npart)

    assert np.all(ad["all", "particle_mass"] == ad["all", "particle_mass"][0])

    left_edge = ds.arr([0.0, 0.0, 0.0], "code_length")
    right_edge = ds.arr([4.0, 4.0, 4.0], "code_length")
    center = 0.5 * (left_edge + right_edge)

    reg = ds.region(center, left_edge, right_edge)

    assert np.all(
        np.logical_and(
            reg["all", "particle_position_x"] <= right_edge[0],
            reg["all", "particle_position_x"] >= left_edge[0],
        )
    )

    assert np.all(
        np.logical_and(
            reg["all", "particle_position_y"] <= right_edge[1],
            reg["all", "particle_position_y"] >= left_edge[1],
        )
    )

    assert np.all(
        np.logical_and(
            reg["all", "particle_position_z"] <= right_edge[2],
            reg["all", "particle_position_z"] >= left_edge[2],
        )
    )


RT_particles = "RT_particles/plt00050"


@requires_ds(RT_particles)
def test_RT_particles():
    ds = data_dir_load(RT_particles)
    assert_equal(str(ds), "plt00050")
    for test in small_patch_amr(ds, _castro_fields):
        test_RT_particles.__name__ = test.description
        yield test


@requires_file(RT_particles)
def test_castro_particle_io():
    ds = data_dir_load(RT_particles)

    grid = ds.index.grids[2]
    npart_grid_2 = 49  # read directly from the header
    assert_equal(grid["all", "particle_position_x"].size, npart_grid_2)
    assert_equal(grid["Tracer", "particle_position_y"].size, npart_grid_2)
    assert_equal(grid["all", "particle_position_y"].size, npart_grid_2)

    ad = ds.all_data()
    npart = 49  # read directly from the header
    assert_equal(ad["all", "particle_velocity_x"].size, npart)
    assert_equal(ad["Tracer", "particle_velocity_y"].size, npart)
    assert_equal(ad["all", "particle_velocity_y"].size, npart)

    left_edge = ds.arr([0.0, 0.0, 0.0], "code_length")
    right_edge = ds.arr([0.25, 1.0, 1.0], "code_length")
    center = 0.5 * (left_edge + right_edge)

    reg = ds.region(center, left_edge, right_edge)

    assert np.all(
        np.logical_and(
            reg["all", "particle_position_x"] <= right_edge[0],
            reg["all", "particle_position_x"] >= left_edge[0],
        )
    )

    assert np.all(
        np.logical_and(
            reg["all", "particle_position_y"] <= right_edge[1],
            reg["all", "particle_position_y"] >= left_edge[1],
        )
    )


langmuir = "LangmuirWave/plt00020_v2"


@requires_ds(langmuir)
def test_langmuir():
    ds = data_dir_load(langmuir)
    assert_equal(str(ds), "plt00020_v2")
    for test in small_patch_amr(
        ds, _warpx_fields, input_center="c", input_weight=("mesh", "Ex")
    ):
        test_langmuir.__name__ = test.description
        yield test


plasma = "PlasmaAcceleration/plt00030_v2"


@requires_ds(plasma)
def test_plasma():
    ds = data_dir_load(plasma)
    assert_equal(str(ds), "plt00030_v2")
    for test in small_patch_amr(
        ds, _warpx_fields, input_center="c", input_weight=("mesh", "Ex")
    ):
        test_plasma.__name__ = test.description
        yield test


beam = "GaussianBeam/plt03008"


@requires_ds(beam)
def test_beam():
    ds = data_dir_load(beam)
    assert_equal(str(ds), "plt03008")
    for param in ("number of boxes", "maximum zones"):
        # PR 2807
        # these parameters are only populated if the config file attached to this
        # dataset is read correctly
        assert param in ds.parameters
    for test in small_patch_amr(
        ds, _warpx_fields, input_center="c", input_weight=("mesh", "Ex")
    ):
        test_beam.__name__ = test.description
        yield test


@requires_file(plasma)
def test_warpx_particle_io():
    ds = data_dir_load(plasma)
    grid = ds.index.grids[0]

    # read directly from the header
    npart0_grid_0 = 344
    npart1_grid_0 = 69632

    assert_equal(grid["particle0", "particle_position_x"].size, npart0_grid_0)
    assert_equal(grid["particle1", "particle_position_y"].size, npart1_grid_0)
    assert_equal(grid["all", "particle_position_z"].size, npart0_grid_0 + npart1_grid_0)

    # read directly from the header
    npart0 = 1360
    npart1 = 802816
    ad = ds.all_data()
    assert_equal(ad["particle0", "particle_velocity_x"].size, npart0)
    assert_equal(ad["particle1", "particle_velocity_y"].size, npart1)
    assert_equal(ad["all", "particle_velocity_z"].size, npart0 + npart1)

    np.all(ad["particle1", "particle_mass"] == ad["particle1", "particle_mass"][0])
    np.all(ad["particle0", "particle_mass"] == ad["particle0", "particle_mass"][0])

    left_edge = ds.arr([-7.5e-5, -7.5e-5, -7.5e-5], "code_length")
    right_edge = ds.arr([2.5e-5, 2.5e-5, 2.5e-5], "code_length")
    center = 0.5 * (left_edge + right_edge)

    reg = ds.region(center, left_edge, right_edge)

    assert np.all(
        np.logical_and(
            reg["all", "particle_position_x"] <= right_edge[0],
            reg["all", "particle_position_x"] >= left_edge[0],
        )
    )

    assert np.all(
        np.logical_and(
            reg["all", "particle_position_y"] <= right_edge[1],
            reg["all", "particle_position_y"] >= left_edge[1],
        )
    )

    assert np.all(
        np.logical_and(
            reg["all", "particle_position_z"] <= right_edge[2],
            reg["all", "particle_position_z"] >= left_edge[2],
        )
    )


_raw_fields = [("raw", "Bx"), ("raw", "Ey"), ("raw", "jz")]

laser = "Laser/plt00015"


@requires_ds(laser)
def test_raw_fields():
    for field in _raw_fields:
        yield GridValuesTest(laser, field)


@requires_file(rt)
def test_OrionDataset():
    assert isinstance(data_dir_load(rt), OrionDataset)


@requires_file(LyA)
def test_NyxDataset():
    assert isinstance(data_dir_load(LyA), NyxDataset)


@requires_file("nyx_small/nyx_small_00000")
def test_NyxDataset_2():
    assert isinstance(data_dir_load("nyx_small/nyx_small_00000"), NyxDataset)


@requires_file(RT_particles)
def test_CastroDataset():
    assert isinstance(data_dir_load(RT_particles), CastroDataset)


@requires_file("castro_sod_x_plt00036")
def test_CastroDataset_2():
    assert isinstance(data_dir_load("castro_sod_x_plt00036"), CastroDataset)


@requires_file("castro_sedov_1d_cyl_plt00150")
def test_CastroDataset_3():
    assert isinstance(data_dir_load("castro_sedov_1d_cyl_plt00150"), CastroDataset)


@requires_file(plasma)
def test_WarpXDataset():
    assert isinstance(data_dir_load(plasma), WarpXDataset)


@disable_dataset_cache
@requires_file(plasma)
def test_magnetic_units():
    ds1 = load(plasma)
    assert_allclose(ds1.magnetic_unit.value, 1.0)
    assert str(ds1.magnetic_unit.units) == "T"
    mag_unit1 = ds1.magnetic_unit.to("code_magnetic")
    assert_allclose(mag_unit1.value, 1.0)
    assert str(mag_unit1.units) == "code_magnetic"
    ds2 = load(plasma, unit_system="cgs")
    assert_allclose(ds2.magnetic_unit.value, 1.0e4)
    assert str(ds2.magnetic_unit.units) == "G"
    mag_unit2 = ds2.magnetic_unit.to("code_magnetic")
    assert_allclose(mag_unit2.value, 1.0)
    assert str(mag_unit2.units) == "code_magnetic"


@requires_ds(laser)
def test_WarpXDataset_2():
    assert isinstance(data_dir_load(laser), WarpXDataset)


@requires_file("plt.Cavity00010")
def test_AMReXDataset():
    ds = data_dir_load("plt.Cavity00010", kwargs={"cparam_filename": "inputs"})
    assert isinstance(ds, AMReXDataset)


@requires_file(rt)
def test_units_override():
    units_override_check(rt)


nyx_no_particles = "nyx_sedov_plt00086"


@requires_file(nyx_no_particles)
def test_nyx_no_part():
    assert isinstance(data_dir_load(nyx_no_particles), NyxDataset)

    fields = sorted(
        [
            ("boxlib", "H"),
            ("boxlib", "He"),
            ("boxlib", "MachNumber"),
            ("boxlib", "Ne"),
            ("boxlib", "Rank"),
            ("boxlib", "StateErr"),
            ("boxlib", "Temp"),
            ("boxlib", "X(H)"),
            ("boxlib", "X(He)"),
            ("boxlib", "density"),
            ("boxlib", "divu"),
            ("boxlib", "eint_E"),
            ("boxlib", "eint_e"),
            ("boxlib", "entropy"),
            ("boxlib", "forcex"),
            ("boxlib", "forcey"),
            ("boxlib", "forcez"),
            ("boxlib", "kineng"),
            ("boxlib", "logden"),
            ("boxlib", "magmom"),
            ("boxlib", "magvel"),
            ("boxlib", "magvort"),
            ("boxlib", "pressure"),
            ("boxlib", "rho_E"),
            ("boxlib", "rho_H"),
            ("boxlib", "rho_He"),
            ("boxlib", "rho_e"),
            ("boxlib", "soundspeed"),
            ("boxlib", "x_velocity"),
            ("boxlib", "xmom"),
            ("boxlib", "y_velocity"),
            ("boxlib", "ymom"),
            ("boxlib", "z_velocity"),
            ("boxlib", "zmom"),
        ]
    )

    ds = data_dir_load(nyx_no_particles)
    assert_equal(sorted(ds.field_list), fields)


msubch = "maestro_subCh_plt00248"


@requires_file(msubch)
def test_maestro_parameters():
    assert isinstance(data_dir_load(msubch), MaestroDataset)
    ds = data_dir_load(msubch)

    # Check a string parameter
    assert ds.parameters["plot_base_name"] == "subCh_hot_baserun_plt"
    assert type(ds.parameters["plot_base_name"]) is str  # noqa: E721

    # Check boolean parameters: T or F
    assert not ds.parameters["use_thermal_diffusion"]
    assert type(ds.parameters["use_thermal_diffusion"]) is bool  # noqa: E721

    assert ds.parameters["do_burning"]
    assert type(ds.parameters["do_burning"]) is bool  # noqa: E721

    # Check a float parameter with a decimal point
    assert ds.parameters["sponge_kappa"] == float("10.00000000")
    assert type(ds.parameters["sponge_kappa"]) is float  # noqa: E721

    # Check a float parameter with E exponent notation
    assert ds.parameters["small_dt"] == float("0.1000000000E-09")

    # Check an int parameter
    assert ds.parameters["s0_interp_type"] == 3
    assert type(ds.parameters["s0_interp_type"]) is int  # noqa: E721


castro_1d_cyl = "castro_sedov_1d_cyl_plt00150"


@requires_file(castro_1d_cyl)
def test_castro_parameters():
    ds = data_dir_load(castro_1d_cyl)
    assert isinstance(ds, CastroDataset)

    # Modified from default (leading [*])
    assert ds.parameters["castro.do_hydro"] == 1
    assert ds.parameters["castro.cfl"] == 0.5
    assert ds.parameters["problem.p_ambient"] == float("1e-06")
    # Leading [*] should be removed from the parameter name
    assert "[*] castro.do_hydro" not in ds.parameters

    # Not modified from default
    assert ds.parameters["castro.pslope_cutoff_density"] == float("-1e+20")
    assert ds.parameters["castro.do_sponge"] == 0
    assert ds.parameters["problem.dens_ambient"] == 1
    assert ds.parameters["eos.eos_assume_neutral"] == 1

    # Empty string value
    assert ds.parameters["castro.stopping_criterion_field"] is None



import os
import tempfile
from unittest import mock

import numpy as np
from numpy.testing import assert_equal

from yt.testing import fake_random_ds
from yt.units.unit_object import Unit


def setup_module():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def teardown_func(fns):
    for fn in fns:
        try:
            os.remove(fn)
        except OSError:
            pass


@mock.patch("matplotlib.backends.backend_agg.FigureCanvasAgg.print_figure")
def test_slice(pf):
    fns = []
    grid_eps = np.finfo(np.float64).eps
    for nprocs in [8, 1]:
        # We want to test both 1 proc and 8 procs, to make sure that
        # parallelism isn't broken
        ds = fake_random_ds(64, nprocs=nprocs)
        dims = ds.domain_dimensions
        xn, yn, zn = ds.domain_dimensions
        dx = ds.arr(1.0 / (ds.domain_dimensions * 2), "code_length")
        xi, yi, zi = ds.domain_left_edge + dx
        xf, yf, zf = ds.domain_right_edge - dx
        coords = np.mgrid[xi : xf : xn * 1j, yi : yf : yn * 1j, zi : zf : zn * 1j]
        uc = [np.unique(c) for c in coords]
        slc_pos = 0.5
        # Some simple slice tests with single grids
        for ax in range(3):
            xax = ds.coordinates.x_axis[ax]
            yax = ds.coordinates.y_axis[ax]
            slc = ds.slice(ax, slc_pos)
            shifted_slc = ds.slice(ax, slc_pos + grid_eps)
            assert_equal(slc["index", "ones"].sum(), slc["index", "ones"].size)
            assert_equal(slc["index", "ones"].min(), 1.0)
            assert_equal(slc["index", "ones"].max(), 1.0)
            assert_equal(np.unique(slc["px"]), uc[xax])
            assert_equal(np.unique(slc["py"]), uc[yax])
            assert_equal(np.unique(slc["pdx"]), 0.5 / dims[xax])
            assert_equal(np.unique(slc["pdy"]), 0.5 / dims[yax])
            pw = slc.to_pw(fields=("gas", "density"))
            for p in pw.plots.values():
                tmpfd, tmpname = tempfile.mkstemp(suffix=".png")
                os.close(tmpfd)
                p.save(name=tmpname)
                fns.append(tmpname)
            for width in [(1.0, "unitary"), 1.0, ds.quan(0.5, "code_length")]:
                frb = slc.to_frb(width, 64)
                shifted_frb = shifted_slc.to_frb(width, 64)
                for slc_field in [("index", "ones"), ("gas", "density")]:
                    fi = ds._get_field_info(slc_field)
                    assert_equal(frb[slc_field].info["data_source"], slc.__str__())
                    assert_equal(frb[slc_field].info["axis"], ax)
                    assert_equal(frb[slc_field].info["field"], str(slc_field))
                    assert_equal(frb[slc_field].units, Unit(fi.units))
                    assert_equal(frb[slc_field].info["xlim"], frb.bounds[:2])
                    assert_equal(frb[slc_field].info["ylim"], frb.bounds[2:])
                    assert_equal(frb[slc_field].info["center"], slc.center)
                    assert_equal(frb[slc_field].info["coord"], slc_pos)
                    assert_equal(frb[slc_field], shifted_frb[slc_field])
    teardown_func(fns)


def test_slice_over_edges():
    ds = fake_random_ds(
        64, nprocs=8, fields=("density",), units=("g/cm**3",), negative=[False]
    )
    slc = ds.slice(0, 0.0)
    slc["gas", "density"]
    slc = ds.slice(1, 0.5)
    slc["gas", "density"]


def test_slice_over_outer_boundary():
    ds = fake_random_ds(
        64, nprocs=8, fields=("density",), units=("g/cm**3",), negative=[False]
    )
    slc = ds.slice(2, 1.0)
    slc["gas", "density"]
    assert_equal(slc["gas", "density"].size, 0)



import warnings
from functools import wraps
from types import FunctionType


def issue_deprecation_warning(
    msg: str,
    *,
    stacklevel: int,
    since: str,
    removal: str | None = None,
):
    """
    Parameters
    ----------
    msg : str
        A text message explaining that the code surrounding the call to this function is
        deprecated, and what should be changed on the user side to avoid it.

    stacklevel: int
        Number of stack frames to be skipped when pointing at caller code, starting from
        *this* function's frame. In general 3 is a minimum.

    since and removal: str version numbers, indicating the anticipated removal date

    Notes
    -----

    removal can be left empty if it is not clear how many minor releases are expected to
    happen before the next major.

    removal and since arguments are keyword-only to forbid accidentally swapping them.

    Examples
    --------
    >>> issue_deprecation_warning(
    ...     "This code is deprecated.", stacklevel=3, since="4.0"
    ... )
    """

    msg += f"\nDeprecated since yt {since}"
    if removal is not None:
        msg += f"\nThis feature is planned for removal in yt {removal}"
    warnings.warn(msg, DeprecationWarning, stacklevel=stacklevel)


def future_positional_only(positions2names: dict[int, str], /, **depr_kwargs):
    """Warn users when using a future positional-only argument as keyword.
    Note that positional-only arguments are available from Python 3.8
    See https://www.python.org/dev/peps/pep-0570/
    """

    def outer(func: FunctionType):
        @wraps(func)
        def inner(*args, **kwargs):
            for no, name in sorted(positions2names.items()):
                if name not in kwargs:
                    continue
                value = kwargs[name]
                issue_deprecation_warning(
                    f"Using the {name!r} argument as keyword (on position {no}) "
                    "is deprecated. "
                    "Pass the argument as positional to suppress this warning, "
                    f"i.e., use {func.__name__}({value!r}, ...)",
                    stacklevel=3,
                    **depr_kwargs,
                )
            return func(*args, **kwargs)

        return inner

    return outer
